@@ -35,17 +35,20 @@
  * is received from the client. It will also call
  * gst_rtsp_stream_transport_set_timed_out() when a receiver has timed out.
  *
+ * A #GstRTSPClient will call gst_rtsp_stream_transport_message_sent() when it
+ * has sent a data message for the transport.
+ *
  * Last reviewed on 2013-07-16 (1.0.0)
  */
+#ifdef HAVE_CONFIG_H
+#include "config.h"
+#endif
 
 #include <string.h>
 #include <stdlib.h>
 
 #include "rtsp-stream-transport.h"
 
-#define GST_RTSP_STREAM_TRANSPORT_GET_PRIVATE(obj)  \
-       (G_TYPE_INSTANCE_GET_PRIVATE ((obj), GST_TYPE_RTSP_STREAM_TRANSPORT, GstRTSPStreamTransportPrivate))
-
 struct _GstRTSPStreamTransportPrivate
 {
   GstRTSPStream *stream;
@@ -61,6 +64,10 @@ struct _GstRTSPStreamTransportPrivate
   gboolean active;
   gboolean timed_out;
 
+  GstRTSPMessageSentFunc message_sent;
+  gpointer ms_user_data;
+  GDestroyNotify ms_notify;
+
   GstRTSPTransport *transport;
   GstRTSPUrl *url;
 
@@ -78,7 +85,7 @@ GST_DEBUG_CATEGORY_STATIC (rtsp_stream_transport_debug);
 
 static void gst_rtsp_stream_transport_finalize (GObject * obj);
 
-G_DEFINE_TYPE (GstRTSPStreamTransport, gst_rtsp_stream_transport,
+G_DEFINE_TYPE_WITH_PRIVATE (GstRTSPStreamTransport, gst_rtsp_stream_transport,
     G_TYPE_OBJECT);
 
 static void
@@ -86,8 +93,6 @@ gst_rtsp_stream_transport_class_init (GstRTSPStreamTransportClass * klass)
 {
   GObjectClass *gobject_class;
 
-  g_type_class_add_private (klass, sizeof (GstRTSPStreamTransportPrivate));
-
   gobject_class = G_OBJECT_CLASS (klass);
 
   gobject_class->finalize = gst_rtsp_stream_transport_finalize;
@@ -99,10 +104,7 @@ gst_rtsp_stream_transport_class_init (GstRTSPStreamTransportClass * klass)
 static void
 gst_rtsp_stream_transport_init (GstRTSPStreamTransport * trans)
 {
-  GstRTSPStreamTransportPrivate *priv =
-      GST_RTSP_STREAM_TRANSPORT_GET_PRIVATE (trans);
-
-  trans->priv = priv;
+  trans->priv = gst_rtsp_stream_transport_get_instance_private (trans);
 }
 
 static void
@@ -117,6 +119,7 @@ gst_rtsp_stream_transport_finalize (GObject * obj)
   /* remove callbacks now */
   gst_rtsp_stream_transport_set_callbacks (trans, NULL, NULL, NULL, NULL);
   gst_rtsp_stream_transport_set_keepalive (trans, NULL, NULL, NULL);
+  gst_rtsp_stream_transport_set_message_sent (trans, NULL, NULL, NULL);
 
   if (priv->stream)
     g_object_unref (priv->stream);
@@ -231,6 +234,33 @@ gst_rtsp_stream_transport_set_keepalive (GstRTSPStreamTransport * trans,
   priv->ka_notify = notify;
 }
 
+/**
+ * gst_rtsp_stream_transport_set_message_sent:
+ * @trans: a #GstRTSPStreamTransport
+ * @message_sent: (scope notified): a callback called when a message has been sent
+ * @user_data: (closure): user data passed to callback
+ * @notify: (allow-none): called with the user_data when no longer needed
+ *
+ * Install a callback that will be called when a message has been sent on @trans.
+ */
+void
+gst_rtsp_stream_transport_set_message_sent (GstRTSPStreamTransport * trans,
+    GstRTSPMessageSentFunc message_sent, gpointer user_data,
+    GDestroyNotify notify)
+{
+  GstRTSPStreamTransportPrivate *priv;
+
+  g_return_if_fail (GST_IS_RTSP_STREAM_TRANSPORT (trans));
+
+  priv = trans->priv;
+
+  priv->message_sent = message_sent;
+  if (priv->ms_notify)
+    priv->ms_notify (priv->ms_user_data);
+  priv->ms_user_data = user_data;
+  priv->ms_notify = notify;
+}
+
 
 /**
  * gst_rtsp_stream_transport_set_transport:
@@ -339,6 +369,8 @@ gst_rtsp_stream_transport_get_rtpinfo (GstRTSPStreamTransport * trans,
 
   priv = trans->priv;
 
+  if (!gst_rtsp_stream_is_sender (priv->stream))
+    return NULL;
   if (!gst_rtsp_stream_get_rtpinfo (priv->stream, &rtptime, &seq, &clock_rate,
           &running_time))
     return NULL;
@@ -516,6 +548,23 @@ gst_rtsp_stream_transport_keep_alive (GstRTSPStreamTransport * trans)
     priv->keep_alive (priv->ka_user_data);
 }
 
+/**
+ * gst_rtsp_stream_transport_message_sent:
+ * @trans: a #GstRTSPStreamTransport
+ *
+ * Signal the installed message_sent callback for @trans.
+ */
+void
+gst_rtsp_stream_transport_message_sent (GstRTSPStreamTransport * trans)
+{
+  GstRTSPStreamTransportPrivate *priv;
+
+  priv = trans->priv;
+
+  if (priv->message_sent)
+    priv->message_sent (priv->ms_user_data);
+}
+
 /**
  * gst_rtsp_stream_transport_recv_data:
  * @trans: a #GstRTSPStreamTransport
