@@ -21,6 +21,13 @@
 
 #include <rtsp-client.h>
 
+#define VIDEO_PIPELINE "videotestsrc ! " \
+  "video/x-raw,width=352,height=288 ! " \
+  "rtpgstpay name=pay0 pt=96"
+#define AUDIO_PIPELINE "audiotestsrc ! " \
+  "audio/x-raw,rate=8000 ! " \
+  "rtpgstpay name=pay1 pt=97"
+
 static gchar *session_id;
 static gint cseq;
 static guint expected_session_timeout = 60;
@@ -135,6 +142,20 @@ test_response_551 (GstRTSPClient * client, GstRTSPMessage * response,
   return TRUE;
 }
 
+static void
+create_connection (GstRTSPConnection ** conn)
+{
+  GSocket *sock;
+  GError *error = NULL;
+
+  sock = g_socket_new (G_SOCKET_FAMILY_IPV4, G_SOCKET_TYPE_STREAM,
+      G_SOCKET_PROTOCOL_TCP, &error);
+  g_assert_no_error (error);
+  fail_unless (gst_rtsp_connection_create_from_socket (sock, "127.0.0.1", 444,
+          NULL, conn) == GST_RTSP_OK);
+  g_object_unref (sock);
+}
+
 static GstRTSPClient *
 setup_client (const gchar * launch_line)
 {
@@ -153,7 +174,7 @@ setup_client (const gchar * launch_line)
   factory = gst_rtsp_media_factory_new ();
   if (launch_line == NULL)
     gst_rtsp_media_factory_set_launch (factory,
-        "videotestsrc ! video/x-raw,width=352,height=288 ! rtpgstpay name=pay0 pt=96");
+        "( " VIDEO_PIPELINE "  " AUDIO_PIPELINE " )");
   else
     gst_rtsp_media_factory_set_launch (factory, launch_line);
 
@@ -291,8 +312,6 @@ GST_START_TEST (test_request)
   GstRTSPMessage request = { 0, };
   gchar *str;
   GstRTSPConnection *conn;
-  GSocket *sock;
-  GError *error = NULL;
 
   client = gst_rtsp_client_new ();
 
@@ -325,13 +344,8 @@ GST_START_TEST (test_request)
 
   /* OPTIONS with an absolute path instead of an absolute url */
   /* set host information */
-  sock = g_socket_new (G_SOCKET_FAMILY_IPV4, G_SOCKET_TYPE_STREAM,
-      G_SOCKET_PROTOCOL_TCP, &error);
-  g_assert_no_error (error);
-  gst_rtsp_connection_create_from_socket (sock, "localhost", 444, NULL, &conn);
+  create_connection (&conn);
   fail_unless (gst_rtsp_client_set_connection (client, conn));
-  g_object_unref (sock);
-
   fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_OPTIONS,
           "/test") == GST_RTSP_OK);
   str = g_strdup_printf ("%d", cseq);
@@ -472,13 +486,14 @@ GST_END_TEST;
 static const gchar *expected_transport = NULL;
 
 static gboolean
-test_setup_response_200_multicast (GstRTSPClient * client,
-    GstRTSPMessage * response, gboolean close, gpointer user_data)
+test_setup_response_200 (GstRTSPClient * client, GstRTSPMessage * response,
+    gboolean close, gpointer user_data)
 {
   GstRTSPStatusCode code;
   const gchar *reason;
   GstRTSPVersion version;
   gchar *str;
+  gchar *pattern;
   GstRTSPSessionPool *session_pool;
   GstRTSPSession *session;
   gchar **session_hdr_params;
@@ -502,7 +517,10 @@ test_setup_response_200_multicast (GstRTSPClient * client,
   fail_unless (gst_rtsp_message_get_header (response, GST_RTSP_HDR_TRANSPORT,
           &str, 0) == GST_RTSP_OK);
 
-  fail_unless_equals_string (str, expected_transport);
+  pattern = g_strdup_printf ("^%s$", expected_transport);
+  fail_unless (g_regex_match_simple (pattern, str, 0, 0),
+      "Transport '%s' doesn't match pattern '%s'", str, pattern);
+  g_free (pattern);
 
   fail_unless (gst_rtsp_message_get_header (response, GST_RTSP_HDR_SESSION,
           &str, 0) == GST_RTSP_OK);
@@ -525,11 +543,12 @@ test_setup_response_200_multicast (GstRTSPClient * client,
   session_pool = gst_rtsp_client_get_session_pool (client);
   fail_unless (session_pool != NULL);
 
-  fail_unless (gst_rtsp_session_pool_get_n_sessions (session_pool) == 1);
   session = gst_rtsp_session_pool_find (session_pool, session_hdr_params[0]);
   g_strfreev (session_hdr_params);
 
   /* remember session id to be able to send teardown */
+  if (session_id)
+    g_free (session_id);
   session_id = g_strdup (gst_rtsp_session_get_sessionid (session));
   fail_unless (session_id != NULL);
 
@@ -542,6 +561,35 @@ test_setup_response_200_multicast (GstRTSPClient * client,
   return TRUE;
 }
 
+static gboolean
+test_setup_response_461 (GstRTSPClient * client,
+    GstRTSPMessage * response, gboolean close, gpointer user_data)
+{
+  GstRTSPStatusCode code;
+  const gchar *reason;
+  GstRTSPVersion version;
+  gchar *str;
+
+  fail_unless (expected_transport == NULL);
+
+  fail_unless (gst_rtsp_message_get_type (response) ==
+      GST_RTSP_MESSAGE_RESPONSE);
+
+  fail_unless (gst_rtsp_message_parse_response (response, &code, &reason,
+          &version)
+      == GST_RTSP_OK);
+  fail_unless (code == GST_RTSP_STS_UNSUPPORTED_TRANSPORT);
+  fail_unless (g_str_equal (reason, "Unsupported transport"));
+  fail_unless (version == GST_RTSP_VERSION_1_0);
+
+  fail_unless (gst_rtsp_message_get_header (response, GST_RTSP_HDR_CSEQ, &str,
+          0) == GST_RTSP_OK);
+  fail_unless (atoi (str) == cseq++);
+
+
+  return TRUE;
+}
+
 static gboolean
 test_teardown_response_200 (GstRTSPClient * client,
     GstRTSPMessage * response, gboolean close, gpointer user_data)
@@ -584,8 +632,90 @@ send_teardown (GstRTSPClient * client)
   session_id = NULL;
 }
 
+GST_START_TEST (test_setup_tcp)
+{
+  GstRTSPClient *client;
+  GstRTSPConnection *conn;
+  GstRTSPMessage request = { 0, };
+  gchar *str;
+
+  client = setup_client (NULL);
+  create_connection (&conn);
+  fail_unless (gst_rtsp_client_set_connection (client, conn));
+
+  fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_SETUP,
+          "rtsp://localhost/test/stream=0") == GST_RTSP_OK);
+  str = g_strdup_printf ("%d", cseq);
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_CSEQ, str);
+  g_free (str);
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_TRANSPORT,
+      "RTP/AVP/TCP;unicast");
+
+  gst_rtsp_client_set_send_func (client, test_setup_response_200, NULL, NULL);
+  expected_transport =
+      "RTP/AVP/TCP;unicast;interleaved=0-1;ssrc=.*;mode=\"PLAY\"";
+  fail_unless (gst_rtsp_client_handle_message (client,
+          &request) == GST_RTSP_OK);
+
+  gst_rtsp_message_unset (&request);
+
+  send_teardown (client);
+  teardown_client (client);
+}
+
+GST_END_TEST;
+
+GST_START_TEST (test_setup_tcp_two_streams_same_channels)
+{
+  GstRTSPClient *client;
+  GstRTSPConnection *conn;
+  GstRTSPMessage request = { 0, };
+  gchar *str;
+
+  client = setup_client (NULL);
+  create_connection (&conn);
+  fail_unless (gst_rtsp_client_set_connection (client, conn));
+
+  /* test SETUP of a video stream with 0-1 as interleaved channels */
+  fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_SETUP,
+          "rtsp://localhost/test/stream=0") == GST_RTSP_OK);
+  str = g_strdup_printf ("%d", cseq);
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_CSEQ, str);
+  g_free (str);
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_TRANSPORT,
+      "RTP/AVP/TCP;unicast;interleaved=0-1");
+  gst_rtsp_client_set_send_func (client, test_setup_response_200, NULL, NULL);
+  expected_transport =
+      "RTP/AVP/TCP;unicast;interleaved=0-1;ssrc=.*;mode=\"PLAY\"";
+  fail_unless (gst_rtsp_client_handle_message (client,
+          &request) == GST_RTSP_OK);
+  gst_rtsp_message_unset (&request);
+
+  /* test SETUP of an audio stream with *the same* interleaved channels.
+   * we expect the server to allocate new channel numbers */
+  fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_SETUP,
+          "rtsp://localhost/test/stream=1") == GST_RTSP_OK);
+  str = g_strdup_printf ("%d", cseq);
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_CSEQ, str);
+  g_free (str);
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_TRANSPORT,
+      "RTP/AVP/TCP;unicast;interleaved=0-1");
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_SESSION, session_id);
+  gst_rtsp_client_set_send_func (client, test_setup_response_200, NULL, NULL);
+  expected_transport =
+      "RTP/AVP/TCP;unicast;interleaved=2-3;ssrc=.*;mode=\"PLAY\"";
+  fail_unless (gst_rtsp_client_handle_message (client,
+          &request) == GST_RTSP_OK);
+  gst_rtsp_message_unset (&request);
+
+  send_teardown (client);
+  teardown_client (client);
+}
+
+GST_END_TEST;
+
 static GstRTSPClient *
-setup_multicast_client (void)
+setup_multicast_client (guint max_ttl)
 {
   GstRTSPClient *client;
   GstRTSPSessionPool *session_pool;
@@ -612,6 +742,7 @@ setup_multicast_client (void)
       "media.factory.construct", G_TYPE_BOOLEAN, TRUE, NULL);
   gst_rtsp_mount_points_add_factory (mount_points, "/test", factory);
   gst_rtsp_client_set_mount_points (client, mount_points);
+  gst_rtsp_media_factory_set_max_mcast_ttl (factory, max_ttl);
 
   thread_pool = gst_rtsp_thread_pool_new ();
   gst_rtsp_client_set_thread_pool (client, thread_pool);
@@ -630,7 +761,7 @@ GST_START_TEST (test_client_multicast_transport_404)
   GstRTSPMessage request = { 0, };
   gchar *str;
 
-  client = setup_multicast_client ();
+  client = setup_multicast_client (1);
 
   /* simple SETUP for non-existing url */
   fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_SETUP,
@@ -663,7 +794,7 @@ GST_START_TEST (test_client_multicast_transport)
   GstRTSPMessage request = { 0, };
   gchar *str;
 
-  client = setup_multicast_client ();
+  client = setup_multicast_client (1);
 
   expected_session_timeout = 20;
   g_signal_connect (G_OBJECT (client), "new-session",
@@ -679,8 +810,7 @@ GST_START_TEST (test_client_multicast_transport)
 
   expected_transport = "RTP/AVP;multicast;destination=233.252.0.1;"
       "ttl=1;port=5000-5001;mode=\"PLAY\"";
-  gst_rtsp_client_set_send_func (client, test_setup_response_200_multicast,
-      NULL, NULL);
+  gst_rtsp_client_set_send_func (client, test_setup_response_200, NULL, NULL);
   fail_unless (gst_rtsp_client_handle_message (client,
           &request) == GST_RTSP_OK);
   gst_rtsp_message_unset (&request);
@@ -700,7 +830,7 @@ GST_START_TEST (test_client_multicast_ignore_transport_specific)
   GstRTSPMessage request = { 0, };
   gchar *str;
 
-  client = setup_multicast_client ();
+  client = setup_multicast_client (1);
 
   /* simple SETUP with a valid URI and multicast and a specific dest,
    * but ignore it  */
@@ -713,8 +843,7 @@ GST_START_TEST (test_client_multicast_ignore_transport_specific)
 
   expected_transport = "RTP/AVP;multicast;destination=233.252.0.1;"
       "ttl=1;port=5000-5001;mode=\"PLAY\"";
-  gst_rtsp_client_set_send_func (client, test_setup_response_200_multicast,
-      NULL, NULL);
+  gst_rtsp_client_set_send_func (client, test_setup_response_200, NULL, NULL);
   fail_unless (gst_rtsp_client_handle_message (client,
           &request) == GST_RTSP_OK);
   gst_rtsp_message_unset (&request);
@@ -727,36 +856,8 @@ GST_START_TEST (test_client_multicast_ignore_transport_specific)
 
 GST_END_TEST;
 
-static gboolean
-test_setup_response_461 (GstRTSPClient * client,
-    GstRTSPMessage * response, gboolean close, gpointer user_data)
-{
-  GstRTSPStatusCode code;
-  const gchar *reason;
-  GstRTSPVersion version;
-  gchar *str;
-
-  fail_unless (expected_transport == NULL);
-
-  fail_unless (gst_rtsp_message_get_type (response) ==
-      GST_RTSP_MESSAGE_RESPONSE);
-
-  fail_unless (gst_rtsp_message_parse_response (response, &code, &reason,
-          &version)
-      == GST_RTSP_OK);
-  fail_unless (code == GST_RTSP_STS_UNSUPPORTED_TRANSPORT);
-  fail_unless (g_str_equal (reason, "Unsupported transport"));
-  fail_unless (version == GST_RTSP_VERSION_1_0);
-
-  fail_unless (gst_rtsp_message_get_header (response, GST_RTSP_HDR_CSEQ, &str,
-          0) == GST_RTSP_OK);
-  fail_unless (atoi (str) == cseq++);
-
-
-  return TRUE;
-}
-
-GST_START_TEST (test_client_multicast_invalid_transport_specific)
+static void
+multicast_transport_specific (void)
 {
   GstRTSPClient *client;
   GstRTSPMessage request = { 0, };
@@ -764,7 +865,7 @@ GST_START_TEST (test_client_multicast_invalid_transport_specific)
   GstRTSPSessionPool *session_pool;
   GstRTSPContext ctx = { NULL };
 
-  client = setup_multicast_client ();
+  client = setup_multicast_client (1);
 
   ctx.client = client;
   ctx.auth = gst_rtsp_auth_new ();
@@ -774,92 +875,7 @@ GST_START_TEST (test_client_multicast_invalid_transport_specific)
       "user", NULL);
   gst_rtsp_context_push_current (&ctx);
 
-  /* simple SETUP with a valid URI and multicast, but an invalid ip */
-  fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_SETUP,
-          "rtsp://localhost/test/stream=0") == GST_RTSP_OK);
-  str = g_strdup_printf ("%d", cseq);
-  gst_rtsp_message_take_header (&request, GST_RTSP_HDR_CSEQ, str);
-  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_TRANSPORT,
-      "RTP/AVP;multicast;destination=233.252.0.2;ttl=1;port=5000-5001;");
-
-  gst_rtsp_client_set_send_func (client, test_setup_response_461, NULL, NULL);
-  fail_unless (gst_rtsp_client_handle_message (client,
-          &request) == GST_RTSP_OK);
-  gst_rtsp_message_unset (&request);
-
-  session_pool = gst_rtsp_client_get_session_pool (client);
-  fail_unless (session_pool != NULL);
-  fail_unless (gst_rtsp_session_pool_get_n_sessions (session_pool) == 0);
-  g_object_unref (session_pool);
-
-
-  /* simple SETUP with a valid URI and multicast, but an invalid prt */
-  fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_SETUP,
-          "rtsp://localhost/test/stream=0") == GST_RTSP_OK);
-  str = g_strdup_printf ("%d", cseq);
-  gst_rtsp_message_take_header (&request, GST_RTSP_HDR_CSEQ, str);
-  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_TRANSPORT,
-      "RTP/AVP;multicast;destination=233.252.0.1;ttl=1;port=6000-6001;");
-
-  gst_rtsp_client_set_send_func (client, test_setup_response_461, NULL, NULL);
-  fail_unless (gst_rtsp_client_handle_message (client,
-          &request) == GST_RTSP_OK);
-  gst_rtsp_message_unset (&request);
-
-  session_pool = gst_rtsp_client_get_session_pool (client);
-  fail_unless (session_pool != NULL);
-  fail_unless (gst_rtsp_session_pool_get_n_sessions (session_pool) == 0);
-  g_object_unref (session_pool);
-
-
-  /* simple SETUP with a valid URI and multicast, but an invalid ttl */
-  fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_SETUP,
-          "rtsp://localhost/test/stream=0") == GST_RTSP_OK);
-  str = g_strdup_printf ("%d", cseq);
-  gst_rtsp_message_take_header (&request, GST_RTSP_HDR_CSEQ, str);
-  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_TRANSPORT,
-      "RTP/AVP;multicast;destination=233.252.0.1;ttl=2;port=5000-5001;");
-
-  gst_rtsp_client_set_send_func (client, test_setup_response_461, NULL, NULL);
-  fail_unless (gst_rtsp_client_handle_message (client,
-          &request) == GST_RTSP_OK);
-  gst_rtsp_message_unset (&request);
-
-  session_pool = gst_rtsp_client_get_session_pool (client);
-  fail_unless (session_pool != NULL);
-  fail_unless (gst_rtsp_session_pool_get_n_sessions (session_pool) == 0);
-  g_object_unref (session_pool);
-
-  teardown_client (client);
-  g_object_unref (ctx.auth);
-  gst_rtsp_token_unref (ctx.token);
-  gst_rtsp_context_pop_current (&ctx);
-}
-
-GST_END_TEST;
-
-GST_START_TEST (test_client_multicast_transport_specific)
-{
-  GstRTSPClient *client;
-  GstRTSPMessage request = { 0, };
-  gchar *str;
-  GstRTSPSessionPool *session_pool;
-  GstRTSPContext ctx = { NULL };
-
-  client = setup_multicast_client ();
-
-  ctx.client = client;
-  ctx.auth = gst_rtsp_auth_new ();
-  ctx.token =
-      gst_rtsp_token_new (GST_RTSP_TOKEN_TRANSPORT_CLIENT_SETTINGS,
-      G_TYPE_BOOLEAN, TRUE, GST_RTSP_TOKEN_MEDIA_FACTORY_ROLE, G_TYPE_STRING,
-      "user", NULL);
-  gst_rtsp_context_push_current (&ctx);
-
-  expected_transport = "RTP/AVP;multicast;destination=233.252.0.1;"
-      "ttl=1;port=5000-5001;mode=\"PLAY\"";
-
-  /* simple SETUP with a valid URI and multicast, but an invalid ip */
+  /* simple SETUP with a valid URI */
   fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_SETUP,
           "rtsp://localhost/test/stream=0") == GST_RTSP_OK);
   str = g_strdup_printf ("%d", cseq);
@@ -867,28 +883,55 @@ GST_START_TEST (test_client_multicast_transport_specific)
   gst_rtsp_message_add_header (&request, GST_RTSP_HDR_TRANSPORT,
       expected_transport);
 
-  gst_rtsp_client_set_send_func (client, test_setup_response_200_multicast,
-      NULL, NULL);
+  gst_rtsp_client_set_send_func (client, test_setup_response_200, NULL, NULL);
   fail_unless (gst_rtsp_client_handle_message (client,
           &request) == GST_RTSP_OK);
   gst_rtsp_message_unset (&request);
-  expected_transport = NULL;
 
-  gst_rtsp_client_set_send_func (client, test_setup_response_200_multicast,
-      NULL, NULL);
+  gst_rtsp_client_set_send_func (client, test_setup_response_200, NULL, NULL);
   session_pool = gst_rtsp_client_get_session_pool (client);
   fail_unless (session_pool != NULL);
   fail_unless (gst_rtsp_session_pool_get_n_sessions (session_pool) == 1);
   g_object_unref (session_pool);
 
-  send_teardown (client);
+  /* send PLAY request */
+  fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_PLAY,
+          "rtsp://localhost/test") == GST_RTSP_OK);
+  str = g_strdup_printf ("%d", cseq);
+  gst_rtsp_message_take_header (&request, GST_RTSP_HDR_CSEQ, str);
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_SESSION, session_id);
+  gst_rtsp_client_set_send_func (client, test_response_200, NULL, NULL);
+  fail_unless (gst_rtsp_client_handle_message (client,
+          &request) == GST_RTSP_OK);
+  gst_rtsp_message_unset (&request);
 
+  send_teardown (client);
   teardown_client (client);
   g_object_unref (ctx.auth);
   gst_rtsp_token_unref (ctx.token);
   gst_rtsp_context_pop_current (&ctx);
 }
 
+/* CASE: multicast address requested by the client exists in the address pool */
+GST_START_TEST (test_client_multicast_transport_specific)
+{
+  expected_transport = "RTP/AVP;multicast;destination=233.252.0.1;"
+      "ttl=1;port=5000-5001;mode=\"PLAY\"";
+  multicast_transport_specific ();
+  expected_transport = NULL;
+}
+
+GST_END_TEST;
+
+/* CASE: multicast address requested by the client does not exist in the address pool */
+GST_START_TEST (test_client_multicast_transport_specific_no_address_in_pool)
+{
+  expected_transport = "RTP/AVP;multicast;destination=234.252.0.3;"
+      "ttl=1;port=10002-10004;mode=\"PLAY\"";
+  multicast_transport_specific ();
+  expected_transport = NULL;
+}
+
 GST_END_TEST;
 
 static gboolean
@@ -1031,6 +1074,436 @@ GST_START_TEST (test_client_sdp_with_no_bitrate_tags)
 
 GST_END_TEST;
 
+static void
+mcast_transport_two_clients (gboolean shared, const gchar * transport1,
+    const gchar * expected_transport1, const gchar * addr1,
+    const gchar * transport2, const gchar * expected_transport2,
+    const gchar * addr2, gboolean bind_mcast_address)
+{
+  GstRTSPClient *client1, *client2;
+  GstRTSPMessage request = { 0, };
+  gchar *str;
+  GstRTSPSessionPool *session_pool;
+  GstRTSPContext ctx = { NULL };
+  GstRTSPContext ctx2 = { NULL };
+  GstRTSPMountPoints *mount_points;
+  GstRTSPMediaFactory *factory;
+  GstRTSPAddressPool *address_pool;
+  GstRTSPThreadPool *thread_pool;
+  gchar *session_id1;
+  gchar *client_addr = NULL;
+
+  mount_points = gst_rtsp_mount_points_new ();
+  factory = gst_rtsp_media_factory_new ();
+  if (shared)
+    gst_rtsp_media_factory_set_shared (factory, TRUE);
+  gst_rtsp_media_factory_set_max_mcast_ttl (factory, 5);
+  gst_rtsp_media_factory_set_bind_mcast_address (factory, bind_mcast_address);
+  gst_rtsp_media_factory_set_launch (factory,
+      "audiotestsrc ! audio/x-raw,rate=44100 ! audioconvert ! rtpL16pay name=pay0");
+  address_pool = gst_rtsp_address_pool_new ();
+  fail_unless (gst_rtsp_address_pool_add_range (address_pool,
+          "233.252.0.1", "233.252.0.1", 5000, 5001, 1));
+  gst_rtsp_media_factory_set_address_pool (factory, address_pool);
+  gst_rtsp_media_factory_add_role (factory, "user",
+      "media.factory.access", G_TYPE_BOOLEAN, TRUE,
+      "media.factory.construct", G_TYPE_BOOLEAN, TRUE, NULL);
+  gst_rtsp_mount_points_add_factory (mount_points, "/test", factory);
+  session_pool = gst_rtsp_session_pool_new ();
+  thread_pool = gst_rtsp_thread_pool_new ();
+
+  /* first multicast client with transport specific request */
+  client1 = gst_rtsp_client_new ();
+  gst_rtsp_client_set_session_pool (client1, session_pool);
+  gst_rtsp_client_set_mount_points (client1, mount_points);
+  gst_rtsp_client_set_thread_pool (client1, thread_pool);
+
+  ctx.client = client1;
+  ctx.auth = gst_rtsp_auth_new ();
+  ctx.token =
+      gst_rtsp_token_new (GST_RTSP_TOKEN_TRANSPORT_CLIENT_SETTINGS,
+      G_TYPE_BOOLEAN, TRUE, GST_RTSP_TOKEN_MEDIA_FACTORY_ROLE, G_TYPE_STRING,
+      "user", NULL);
+  gst_rtsp_context_push_current (&ctx);
+
+  expected_transport = expected_transport1;
+
+  /* send SETUP request */
+  fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_SETUP,
+          "rtsp://localhost/test/stream=0") == GST_RTSP_OK);
+  str = g_strdup_printf ("%d", cseq);
+  gst_rtsp_message_take_header (&request, GST_RTSP_HDR_CSEQ, str);
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_TRANSPORT, transport1);
+
+  gst_rtsp_client_set_send_func (client1, test_setup_response_200, NULL, NULL);
+  fail_unless (gst_rtsp_client_handle_message (client1,
+          &request) == GST_RTSP_OK);
+  gst_rtsp_message_unset (&request);
+  expected_transport = NULL;
+
+  /* send PLAY request */
+  fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_PLAY,
+          "rtsp://localhost/test") == GST_RTSP_OK);
+  str = g_strdup_printf ("%d", cseq);
+  gst_rtsp_message_take_header (&request, GST_RTSP_HDR_CSEQ, str);
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_SESSION, session_id);
+  gst_rtsp_client_set_send_func (client1, test_response_200, NULL, NULL);
+  fail_unless (gst_rtsp_client_handle_message (client1,
+          &request) == GST_RTSP_OK);
+  gst_rtsp_message_unset (&request);
+
+  /* check address */
+  client_addr = gst_rtsp_stream_get_multicast_client_addresses (ctx.stream);
+  fail_if (client_addr == NULL);
+  fail_unless (g_str_equal (client_addr, addr1));
+  g_free (client_addr);
+
+  gst_rtsp_context_pop_current (&ctx);
+  session_id1 = g_strdup (session_id);
+
+  /* second multicast client with transport specific request */
+  cseq = 0;
+  client2 = gst_rtsp_client_new ();
+  gst_rtsp_client_set_session_pool (client2, session_pool);
+  gst_rtsp_client_set_mount_points (client2, mount_points);
+  gst_rtsp_client_set_thread_pool (client2, thread_pool);
+
+  ctx2.client = client2;
+  ctx2.auth = gst_rtsp_auth_new ();
+  ctx2.token =
+      gst_rtsp_token_new (GST_RTSP_TOKEN_TRANSPORT_CLIENT_SETTINGS,
+      G_TYPE_BOOLEAN, TRUE, GST_RTSP_TOKEN_MEDIA_FACTORY_ROLE, G_TYPE_STRING,
+      "user", NULL);
+  gst_rtsp_context_push_current (&ctx2);
+
+  expected_transport = expected_transport2;
+
+  /* send SETUP request */
+  fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_SETUP,
+          "rtsp://localhost/test/stream=0") == GST_RTSP_OK);
+  str = g_strdup_printf ("%d", cseq);
+  gst_rtsp_message_take_header (&request, GST_RTSP_HDR_CSEQ, str);
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_TRANSPORT, transport2);
+
+  gst_rtsp_client_set_send_func (client2, test_setup_response_200, NULL, NULL);
+  fail_unless (gst_rtsp_client_handle_message (client2,
+          &request) == GST_RTSP_OK);
+  gst_rtsp_message_unset (&request);
+  expected_transport = NULL;
+
+  /* send PLAY request */
+  fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_PLAY,
+          "rtsp://localhost/test") == GST_RTSP_OK);
+  str = g_strdup_printf ("%d", cseq);
+  gst_rtsp_message_take_header (&request, GST_RTSP_HDR_CSEQ, str);
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_SESSION, session_id);
+  gst_rtsp_client_set_send_func (client2, test_response_200, NULL, NULL);
+  fail_unless (gst_rtsp_client_handle_message (client2,
+          &request) == GST_RTSP_OK);
+  gst_rtsp_message_unset (&request);
+
+  /* check addresses */
+  client_addr = gst_rtsp_stream_get_multicast_client_addresses (ctx2.stream);
+  fail_if (client_addr == NULL);
+  if (shared) {
+    if (g_str_equal (addr1, addr2)) {
+      fail_unless (g_str_equal (client_addr, addr1));
+    } else {
+      gchar *addr_str = g_strdup_printf ("%s,%s", addr2, addr1);
+      fail_unless (g_str_equal (client_addr, addr_str));
+      g_free (addr_str);
+    }
+  } else {
+    fail_unless (g_str_equal (client_addr, addr2));
+  }
+  g_free (client_addr);
+
+  send_teardown (client2);
+  gst_rtsp_context_pop_current (&ctx2);
+
+  gst_rtsp_context_push_current (&ctx);
+  session_id = session_id1;
+  send_teardown (client1);
+  gst_rtsp_context_pop_current (&ctx);
+
+  teardown_client (client1);
+  teardown_client (client2);
+  g_object_unref (ctx.auth);
+  g_object_unref (ctx2.auth);
+  gst_rtsp_token_unref (ctx.token);
+  gst_rtsp_token_unref (ctx2.token);
+  g_object_unref (mount_points);
+  g_object_unref (session_pool);
+  g_object_unref (address_pool);
+  g_object_unref (thread_pool);
+}
+
+/* test if two multicast clients can choose different transport settings
+ * CASE: media is shared */
+GST_START_TEST
+    (test_client_multicast_transport_specific_two_clients_shared_media) {
+  const gchar *transport_client_1 = "RTP/AVP;multicast;destination=233.252.0.1;"
+      "ttl=1;port=5000-5001;mode=\"PLAY\"";
+  const gchar *expected_transport_1 = transport_client_1;
+  const gchar *addr_client_1 = "233.252.0.1:5000";
+
+  const gchar *transport_client_2 = "RTP/AVP;multicast;destination=233.252.0.2;"
+      "ttl=1;port=5002-5003;mode=\"PLAY\"";
+  const gchar *expected_transport_2 = transport_client_2;
+  const gchar *addr_client_2 = "233.252.0.2:5002";
+
+  mcast_transport_two_clients (TRUE, transport_client_1,
+      expected_transport_1, addr_client_1, transport_client_2,
+      expected_transport_2, addr_client_2, FALSE);
+}
+
+GST_END_TEST;
+
+/* test if two multicast clients can choose different transport settings
+ * CASE: media is not shared */
+GST_START_TEST (test_client_multicast_transport_specific_two_clients)
+{
+  const gchar *transport_client_1 = "RTP/AVP;multicast;destination=233.252.0.1;"
+      "ttl=1;port=5000-5001;mode=\"PLAY\"";
+  const gchar *expected_transport_1 = transport_client_1;
+  const gchar *addr_client_1 = "233.252.0.1:5000";
+
+  const gchar *transport_client_2 = "RTP/AVP;multicast;destination=233.252.0.2;"
+      "ttl=1;port=5002-5003;mode=\"PLAY\"";
+  const gchar *expected_transport_2 = transport_client_2;
+  const gchar *addr_client_2 = "233.252.0.2:5002";
+
+  mcast_transport_two_clients (FALSE, transport_client_1,
+      expected_transport_1, addr_client_1, transport_client_2,
+      expected_transport_2, addr_client_2, FALSE);
+}
+
+GST_END_TEST;
+
+/* test if two multicast clients can choose the same ports but different
+ * multicast destinations
+ * CASE: media is not shared */
+GST_START_TEST (test_client_multicast_transport_specific_two_clients_same_ports)
+{
+  const gchar *transport_client_1 = "RTP/AVP;multicast;destination=233.252.0.1;"
+      "ttl=1;port=9000-9001;mode=\"PLAY\"";
+  const gchar *expected_transport_1 = transport_client_1;
+  const gchar *addr_client_1 = "233.252.0.1:9000";
+
+  const gchar *transport_client_2 = "RTP/AVP;multicast;destination=233.252.0.2;"
+      "ttl=1;port=9000-9001;mode=\"PLAY\"";
+  const gchar *expected_transport_2 = transport_client_2;
+  const gchar *addr_client_2 = "233.252.0.2:9000";
+
+  /* configure the multicast socket to be bound to the requested multicast address instead of INADDR_ANY.
+   * The clients request the same rtp/rtcp borts and having the socket that are bound to ANY would result
+   * in bind() failure */
+  gboolean allow_bind_mcast_address = TRUE;
+
+  mcast_transport_two_clients (FALSE, transport_client_1,
+      expected_transport_1, addr_client_1, transport_client_2,
+      expected_transport_2, addr_client_2, allow_bind_mcast_address);
+}
+
+GST_END_TEST;
+
+/* test if two multicast clients can choose the same multicast destination but different
+ * ports
+ * CASE: media is not shared */
+GST_START_TEST
+    (test_client_multicast_transport_specific_two_clients_same_destination) {
+  const gchar *transport_client_1 = "RTP/AVP;multicast;destination=233.252.0.2;"
+      "ttl=1;port=9002-9003;mode=\"PLAY\"";
+  const gchar *expected_transport_1 = transport_client_1;
+  const gchar *addr_client_1 = "233.252.0.2:9002";
+
+  const gchar *transport_client_2 = "RTP/AVP;multicast;destination=233.252.0.2;"
+      "ttl=1;port=9004-9005;mode=\"PLAY\"";
+  const gchar *expected_transport_2 = transport_client_2;
+  const gchar *addr_client_2 = "233.252.0.2:9004";
+
+  mcast_transport_two_clients (FALSE, transport_client_1,
+      expected_transport_1, addr_client_1, transport_client_2,
+      expected_transport_2, addr_client_2, FALSE);
+}
+
+GST_END_TEST;
+/* test if two multicast clients can choose the same transport settings.
+ * CASE: media is shared */
+GST_START_TEST
+    (test_client_multicast_transport_specific_two_clients_shared_media_same_transport)
+{
+
+  const gchar *transport_client_1 = "RTP/AVP;multicast;destination=233.252.0.1;"
+      "ttl=1;port=5000-5001;mode=\"PLAY\"";
+  const gchar *expected_transport_1 = transport_client_1;
+  const gchar *addr_client_1 = "233.252.0.1:5000";
+
+  const gchar *transport_client_2 = transport_client_1;
+  const gchar *expected_transport_2 = expected_transport_1;
+  const gchar *addr_client_2 = addr_client_1;
+
+  mcast_transport_two_clients (TRUE, transport_client_1,
+      expected_transport_1, addr_client_1, transport_client_2,
+      expected_transport_2, addr_client_2, FALSE);
+}
+
+GST_END_TEST;
+
+/* test if two multicast clients get the same transport settings without
+ * requesting specific transport.
+ * CASE: media is shared */
+GST_START_TEST (test_client_multicast_two_clients_shared_media)
+{
+  const gchar *transport_client_1 = "RTP/AVP;multicast;mode=\"PLAY\"";
+  const gchar *expected_transport_1 =
+      "RTP/AVP;multicast;destination=233.252.0.1;"
+      "ttl=1;port=5000-5001;mode=\"PLAY\"";
+  const gchar *addr_client_1 = "233.252.0.1:5000";
+
+  const gchar *transport_client_2 = transport_client_1;
+  const gchar *expected_transport_2 = expected_transport_1;
+  const gchar *addr_client_2 = addr_client_1;
+
+  mcast_transport_two_clients (TRUE, transport_client_1,
+      expected_transport_1, addr_client_1, transport_client_2,
+      expected_transport_2, addr_client_2, FALSE);
+}
+
+GST_END_TEST;
+
+/* test if two multicast clients get the different transport settings: the first client 
+ * requests the specific transport configuration while the second client lets
+ * the server select the multicast address and the ports.
+ * CASE: media is shared */
+GST_START_TEST
+    (test_client_multicast_two_clients_first_specific_transport_shared_media) {
+  const gchar *transport_client_1 = "RTP/AVP;multicast;destination=233.252.0.1;"
+      "ttl=1;port=5000-5001;mode=\"PLAY\"";
+  const gchar *expected_transport_1 = transport_client_1;
+  const gchar *addr_client_1 = "233.252.0.1:5000";
+
+  const gchar *transport_client_2 = "RTP/AVP;multicast;mode=\"PLAY\"";
+  const gchar *expected_transport_2 = expected_transport_1;
+  const gchar *addr_client_2 = addr_client_1;
+
+  mcast_transport_two_clients (TRUE, transport_client_1,
+      expected_transport_1, addr_client_1, transport_client_2,
+      expected_transport_2, addr_client_2, FALSE);
+}
+
+GST_END_TEST;
+/* test if two multicast clients get the different transport settings: the first client lets
+ * the server select the multicast address and the ports while the second client requests 
+ * the specific transport configuration.
+ * CASE: media is shared */
+GST_START_TEST
+    (test_client_multicast_two_clients_second_specific_transport_shared_media) {
+  const gchar *transport_client_1 = "RTP/AVP;multicast;mode=\"PLAY\"";
+  const gchar *expected_transport_1 =
+      "RTP/AVP;multicast;destination=233.252.0.1;"
+      "ttl=1;port=5000-5001;mode=\"PLAY\"";
+  const gchar *addr_client_1 = "233.252.0.1:5000";
+
+  const gchar *transport_client_2 = "RTP/AVP;multicast;destination=233.252.0.2;"
+      "ttl=2;port=5004-5005;mode=\"PLAY\"";
+  const gchar *expected_transport_2 = transport_client_2;
+  const gchar *addr_client_2 = "233.252.0.2:5004";
+
+  mcast_transport_two_clients (TRUE, transport_client_1,
+      expected_transport_1, addr_client_1, transport_client_2,
+      expected_transport_2, addr_client_2, FALSE);
+}
+
+GST_END_TEST;
+
+/* test if the maximum ttl multicast value is chosen by the server
+ * CASE: the first client provides the highest ttl value */
+GST_START_TEST (test_client_multicast_max_ttl_first_client)
+{
+  const gchar *transport_client_1 = "RTP/AVP;multicast;destination=233.252.0.1;"
+      "ttl=3;port=5000-5001;mode=\"PLAY\"";
+  const gchar *expected_transport_1 = transport_client_1;
+  const gchar *addr_client_1 = "233.252.0.1:5000";
+
+  const gchar *transport_client_2 = "RTP/AVP;multicast;destination=233.252.0.2;"
+      "ttl=1;port=5002-5003;mode=\"PLAY\"";
+  const gchar *expected_transport_2 =
+      "RTP/AVP;multicast;destination=233.252.0.2;"
+      "ttl=3;port=5002-5003;mode=\"PLAY\"";
+  const gchar *addr_client_2 = "233.252.0.2:5002";
+
+  mcast_transport_two_clients (TRUE, transport_client_1,
+      expected_transport_1, addr_client_1, transport_client_2,
+      expected_transport_2, addr_client_2, FALSE);
+}
+
+GST_END_TEST;
+
+/* test if the maximum ttl multicast value is chosen by the server
+ * CASE: the second client provides the highest ttl value */
+GST_START_TEST (test_client_multicast_max_ttl_second_client)
+{
+  const gchar *transport_client_1 = "RTP/AVP;multicast;destination=233.252.0.1;"
+      "ttl=2;port=5000-5001;mode=\"PLAY\"";
+  const gchar *expected_transport_1 = transport_client_1;
+  const gchar *addr_client_1 = "233.252.0.1:5000";
+
+  const gchar *transport_client_2 = "RTP/AVP;multicast;destination=233.252.0.2;"
+      "ttl=4;port=5002-5003;mode=\"PLAY\"";
+  const gchar *expected_transport_2 = transport_client_2;
+  const gchar *addr_client_2 = "233.252.0.2:5002";
+
+  mcast_transport_two_clients (TRUE, transport_client_1,
+      expected_transport_1, addr_client_1, transport_client_2,
+      expected_transport_2, addr_client_2, FALSE);
+}
+
+GST_END_TEST;
+GST_START_TEST (test_client_multicast_invalid_ttl)
+{
+  GstRTSPClient *client;
+  GstRTSPMessage request = { 0, };
+  gchar *str;
+  GstRTSPSessionPool *session_pool;
+  GstRTSPContext ctx = { NULL };
+
+  client = setup_multicast_client (3);
+
+  ctx.client = client;
+  ctx.auth = gst_rtsp_auth_new ();
+  ctx.token =
+      gst_rtsp_token_new (GST_RTSP_TOKEN_TRANSPORT_CLIENT_SETTINGS,
+      G_TYPE_BOOLEAN, TRUE, GST_RTSP_TOKEN_MEDIA_FACTORY_ROLE, G_TYPE_STRING,
+      "user", NULL);
+  gst_rtsp_context_push_current (&ctx);
+
+  /* simple SETUP with an invalid ttl=0 */
+  fail_unless (gst_rtsp_message_init_request (&request, GST_RTSP_SETUP,
+          "rtsp://localhost/test/stream=0") == GST_RTSP_OK);
+  str = g_strdup_printf ("%d", cseq);
+  gst_rtsp_message_take_header (&request, GST_RTSP_HDR_CSEQ, str);
+  gst_rtsp_message_add_header (&request, GST_RTSP_HDR_TRANSPORT,
+      "RTP/AVP;multicast;destination=233.252.0.1;ttl=0;port=5000-5001;");
+
+  gst_rtsp_client_set_send_func (client, test_setup_response_461, NULL, NULL);
+  fail_unless (gst_rtsp_client_handle_message (client,
+          &request) == GST_RTSP_OK);
+  gst_rtsp_message_unset (&request);
+
+  session_pool = gst_rtsp_client_get_session_pool (client);
+  fail_unless (session_pool != NULL);
+  fail_unless (gst_rtsp_session_pool_get_n_sessions (session_pool) == 0);
+  g_object_unref (session_pool);
+
+  teardown_client (client);
+  g_object_unref (ctx.auth);
+  gst_rtsp_token_unref (ctx.token);
+  gst_rtsp_context_pop_current (&ctx);
+}
+
+GST_END_TEST;
+
 static Suite *
 rtspclient_suite (void)
 {
@@ -1043,15 +1516,42 @@ rtspclient_suite (void)
   tcase_add_test (tc, test_request);
   tcase_add_test (tc, test_options);
   tcase_add_test (tc, test_describe);
+  tcase_add_test (tc, test_setup_tcp);
+  tcase_add_test (tc, test_setup_tcp_two_streams_same_channels);
   tcase_add_test (tc, test_client_multicast_transport_404);
   tcase_add_test (tc, test_client_multicast_transport);
   tcase_add_test (tc, test_client_multicast_ignore_transport_specific);
-  tcase_add_test (tc, test_client_multicast_invalid_transport_specific);
   tcase_add_test (tc, test_client_multicast_transport_specific);
   tcase_add_test (tc, test_client_sdp_with_max_bitrate_tag);
   tcase_add_test (tc, test_client_sdp_with_bitrate_tag);
   tcase_add_test (tc, test_client_sdp_with_max_bitrate_and_bitrate_tags);
   tcase_add_test (tc, test_client_sdp_with_no_bitrate_tags);
+  tcase_add_test (tc,
+      test_client_multicast_transport_specific_two_clients_shared_media);
+  tcase_add_test (tc, test_client_multicast_transport_specific_two_clients);
+#ifndef G_OS_WIN32
+  tcase_add_test (tc,
+      test_client_multicast_transport_specific_two_clients_same_ports);
+#else
+  /* skip the test on windows as the test restricts the multicast sockets to multicast traffic only,
+   * by specifying the multicast IP as the bind address and this currently doesn't work on Windows */
+  tcase_skip_broken_test (tc,
+      test_client_multicast_transport_specific_two_clients_same_ports);
+#endif
+  tcase_add_test (tc,
+      test_client_multicast_transport_specific_two_clients_same_destination);
+  tcase_add_test (tc,
+      test_client_multicast_transport_specific_two_clients_shared_media_same_transport);
+  tcase_add_test (tc, test_client_multicast_two_clients_shared_media);
+  tcase_add_test (tc,
+      test_client_multicast_two_clients_first_specific_transport_shared_media);
+  tcase_add_test (tc,
+      test_client_multicast_two_clients_second_specific_transport_shared_media);
+  tcase_add_test (tc,
+      test_client_multicast_transport_specific_no_address_in_pool);
+  tcase_add_test (tc, test_client_multicast_max_ttl_first_client);
+  tcase_add_test (tc, test_client_multicast_max_ttl_second_client);
+  tcase_add_test (tc, test_client_multicast_invalid_ttl);
 
   return s;
 }
